/*****************************************************************************
 * Filename:    cryptfs/vt.c                                                 *
 * Description: TTY handling                                                 *
 * Copyright:   2005 by Alexander Motzkau                                    *
 *****************************************************************************/

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/vt.h>

#include "misc.h"

int min_vt = -1;

int vtnr = -1;
int origvtnr;

void reserve_tty()
{
    struct stat info;
    int fd;
    
    for(fd=0; fd<3; fd++)
	if(fstat(fd,&info)<0)
	{
	    int nfd;
	    nfd = open("/dev/null", O_RDWR);

	    if(nfd<0) /* We're screwed. */
		continue;
		
	    if(nfd==fd)
		continue;
	    
	    dup2(nfd, fd);
	    if(nfd > 2)
		close(nfd);
	}
}


char *devs[] = { "/dev/tty", "/dev/tty0", "/dev/console", NULL};
int modes[] = { O_RDWR, O_RDONLY, O_WRONLY, 0 };

void get_tty()
{
    int fd;
    /* Get's a new tty, if we aren't connected to one. */
    
    if(vtnr>0 || 
	(isatty(STDIN_FILENO) && isatty(STDOUT_FILENO) && isatty(STDERR_FILENO)))
	return;
	
    fd = open("/dev/tty", O_RDWR);
    if(fd < 0)
    {
	/* Let's search for a console to control the virtual terminals. */
	char **tdev;
	int cfd;
	struct vt_stat vtstat;
	
	cfd = -1;
	for(tdev = devs; *tdev != NULL; tdev++)
	{
	    int *mode;
	    for(mode = modes; *mode != 0; mode++)
	    {
		cfd = open(*tdev, *mode);
		if(cfd >= 0 || errno != EACCES)
		    break;
	    }
	    
	    if(ioctl(cfd, VT_GETSTATE, &vtstat) < 0)
		close(cfd);
	    else
		break;
	}
	
	if(*tdev == NULL)
	    return; /* Nothing found. I'm sorry. */
	
	if(ioctl(cfd, VT_OPENQRY, &vtnr) < 0 || vtnr < 0)
	{
	    close(cfd);
	    return;
	}

	if(vtnr < min_vt)
	    vtnr = min_vt;
	
	do
	{
	    char vtname[32];
	    
	    if(vtnr<16 && (vtstat.v_state & (1<<vtnr)))
	    {
		/* Already in use. */
		vtnr++;
		continue;
	    }
	    
	    if(snprintf(vtname, 32, "/dev/tty%d", vtnr) > 31)
	    {
		close(cfd);
		return;
	    }
	    
	    fd = open(vtname, O_RDWR);
	    if(fd >= 0)
		break;
	    
	    vtnr++;
	}while(vtnr < 16);
	
	if(fd < 0)
	{
	    close(cfd);
	    errmsg("Couldn't find a tty.\n");
	    return;
	}
	
	ioctl(cfd, VT_ACTIVATE, vtnr);
	ioctl(cfd, VT_WAITACTIVE, vtnr);
	close(cfd);
	origvtnr = vtstat.v_active;

	setenv("TERM", "linux", 1);
    }

    if(dup2(fd, STDIN_FILENO)<0) errnomsg("dup2 to stdin");
    if(dup2(fd, STDERR_FILENO)<0) errnomsg("dup2 to stderr");
    if(dup2(fd, STDOUT_FILENO)<0) errnomsg("dup2 to stdout");

    close(fd);
}

void release_tty()
{
    /* Switch back to a virtual terminal, if we switched. */
    struct vt_stat vtstat;
    if(vtnr < 0)
	return;

    if(ioctl(STDERR_FILENO, VT_GETSTATE, &vtstat) < 0)
	return;
    
    if(vtstat.v_active != vtnr)
	return;

    ioctl(STDERR_FILENO, VT_ACTIVATE, origvtnr);
    ioctl(STDERR_FILENO, VT_WAITACTIVE, origvtnr);
}
